from .figure import Figure


def subplot_grid(*args, **kwargs):
    return SubplotGrid(*args, **kwargs)


class SubplotGrid:
    def __init__(
        self,
        layout,
        width=None,
        column_widths=None,
        border_style="thin",
        padding=(1, 2),
    ):
        assert (
            len(layout) == 2
        ), "layout must be an interable of length 2 (rows and columns)"

        self._layout = layout

        if border_style is None:
            self._border_chars = None
        elif len(border_style) == 1:
            self._border_chars = 11 * [border_style]
        elif isinstance(border_style, list):
            assert len(border_style) == 11
            self._border_chars = border_style
        else:
            self._border_chars = {
                "thin": ["─", "│", "┌", "┐", "└", "┘", "├", "┤", "┬", "┴", "┼"],
                "thin rounded": ["─", "│", "╭", "╮", "╰", "╯", "├", "┤", "┬", "┴", "┼"],
                "thick": ["━", "┃", "┏", "┓", "┗", "┛", "┣", "┫", "┳", "┻", "╋"],
                "double": ["═", "║", "╔", "╗", "╚", "╝", "╠", "╣", "╦", "╩", "╬"],
                "ascii": ["-", "|", "-", "-", "-", "-", "|", "|", "-", "-", "+"],
            }[border_style]

        border_width = 1
        self._num_borders = layout[1] + 1

        if column_widths is None:
            if width is None:
                self._column_widths = [None] * layout[1]
            else:
                self._column_widths = [
                    (width - self._num_borders * border_width) // layout[1]
                    for _ in range(layout[1])
                ]
                for k in range((width - self._num_borders * border_width) % layout[1]):
                    self._column_widths[k] += 1
        else:
            assert (
                width is None
            ), "At most one of `width` and `column_widths` can be specified."
            assert len(column_widths) == layout[1]
            self._column_widths = column_widths

        self._subfigures = [
            [Figure(self._column_widths[j], padding=padding) for j in range(layout[1])]
            for _ in range(layout[0])
        ]
        return

    def show(self):
        print(self.get_string())
        return

    def get_string(self):
        # compute column width
        cstrings = [
            [item.get_string(remove_trailing_whitespace=False) for item in row]
            for row in self._subfigures
        ]
        column_widths = [
            max(
                [
                    len(line)
                    for i in range(self._layout[0])
                    for line in cstrings[i][j].split("\n")
                ]
            )
            for j in range(self._layout[1])
        ]

        string = []

        if self._border_chars:
            string += [
                self._border_chars[2]
                + self._border_chars[8].join(
                    [s * self._border_chars[0] for s in column_widths]
                )
                + self._border_chars[3]
            ]

        # collect the subfigure rows
        srows = []
        for row in self._subfigures:
            cstrings = [item.get_string().split("\n") for item in row]
            max_num_lines = max(len(item) for item in cstrings)
            pp = []
            for k in range(max_num_lines):
                p = []
                for j, cstring in enumerate(cstrings):
                    try:
                        s = cstring[k]
                    except IndexError:
                        s = ""
                    # truncate or extend with spaces to match the column width
                    if len(s) >= column_widths[j]:
                        s = s[: column_widths[j]]
                    else:
                        s += " " * (column_widths[j] - len(s))
                    p.append(s)
                if self._border_chars:
                    join_char = self._border_chars[1]
                else:
                    join_char = ""
                pp.append(join_char + join_char.join(p) + join_char)
            srows.append("\n".join([p.rstrip() for p in pp]))

        if self._border_chars:
            intermediate_border_row = (
                "\n"
                + self._border_chars[6]
                + self._border_chars[10].join(
                    [s * self._border_chars[0] for s in column_widths]
                )
                + self._border_chars[7]
                + "\n"
            )
        else:
            intermediate_border_row = "\n"
        string += [intermediate_border_row.join(srows)]

        if self._border_chars:
            # final row
            string += [
                self._border_chars[4]
                + self._border_chars[9].join(
                    [s * self._border_chars[0] for s in column_widths]
                )
                + self._border_chars[5]
            ]

        return "\n".join([s.rstrip() for s in string])

    def __getitem__(self, ij):
        i, j = ij
        return self._subfigures[i][j]



import termplotlib as tpl


def test_simple():
    fig = tpl.figure()
    fig.aprint("abc")
    string = fig.get_string()
    assert string == """abc"""
    return


def test_padding_1():
    fig = tpl.figure(padding=1)
    fig.aprint("abc")
    string = fig.get_string()
    assert (
        string
        == """
 abc
"""
    )
    return


def test_padding_1b():
    fig = tpl.figure(padding=(1,))
    fig.aprint("abc")
    string = fig.get_string()
    assert (
        string
        == """
 abc
"""
    )
    return


def test_padding_2():
    fig = tpl.figure(padding=(1, 2))
    fig.aprint("abc")
    string = fig.get_string()
    assert (
        string
        == """
  abc
"""
    )
    return


def test_padding_3():
    fig = tpl.figure(padding=(1, 2, 3))
    fig.aprint("abc")
    string = fig.get_string()
    assert (
        string
        == """
  abc


"""
    )
    return


def test_padding_4():
    fig = tpl.figure(padding=(1, 2, 3, 4))
    fig.aprint("abc")
    string = fig.get_string()
    assert (
        string
        == """
    abc


"""
    )
    fig.show()
    return



from .barh import _get_matrix_of_eighths, _trim_trailing_zeros, barh
from .helpers import is_unicode_standard_output


def hist(
    counts,
    bin_edges,
    orientation="vertical",
    max_width=40,
    bins=20,
    grid=None,
    bar_width=1,
    strip=False,
    force_ascii=False,
):
    if orientation == "vertical":
        return hist_vertical(
            counts,
            xgrid=grid,
            bar_width=bar_width,
            strip=strip,
            force_ascii=force_ascii,
        )

    assert orientation == "horizontal", f"Unknown orientation '{orientation}'"
    return hist_horizontal(
        counts,
        bin_edges,
        max_width=40,
        bins=20,
        bar_width=bar_width,
        force_ascii=force_ascii,
    )


def hist_horizontal(
    counts,
    bin_edges,
    max_width=40,
    bins=20,
    bar_width=1,
    show_bin_edges=True,
    show_counts=True,
    force_ascii=False,
):
    if show_bin_edges:
        labels = [
            "{:+.2e} - {:+.2e}".format(bin_edges[k], bin_edges[k + 1])
            for k in range(len(bin_edges) - 1)
        ]
    else:
        labels = None

    out = barh(
        counts,
        labels=labels,
        max_width=max_width,
        bar_width=bar_width,
        show_vals=show_counts,
        force_ascii=force_ascii,
    )

    return out


def _flip(matrix):
    n = len(matrix[0])
    return [[row[-(i + 1)] for row in matrix] for i in range(n)]


def hist_vertical(
    counts,
    bins=30,
    max_height=10,
    bar_width=2,
    strip=False,
    xgrid=None,
    force_ascii=False,
):
    if xgrid is None:
        xgrid = []

    matrix = _get_matrix_of_eighths(counts, max_height, bar_width)

    if strip:
        # Cut off leading and trailing rows of 0
        k0 = 0
        for row in matrix:
            if any([item != 0 for item in row]):
                break
            k0 += 1

        k1 = 0
        for row in matrix[::-1]:
            if any([item != 0 for item in row]):
                break
            k1 += 1
        n = len(matrix)
        matrix = matrix[k0 : n - k1]
    else:
        k0 = 0

    if is_unicode_standard_output() and not force_ascii:
        block_chars = [" ", "▁", "▂", "▃", "▄", "▅", "▆", "▇", "█"]
        left_seven_eighths = "▉"
    else:
        block_chars = [" ", "*", "*", "*", "*", "*", "*", "*", "*"]
        left_seven_eighths = "*"

    # print text matrix
    out = []
    for row in _flip(matrix):

        # Cut off trailing zeros
        r = _trim_trailing_zeros(row)

        c = [block_chars[item] for item in r]

        # add grid lines
        for i in xgrid:
            # print(row[xgrid])
            pos = (i - k0) * bar_width - 1
            if row[pos] == 8 and (pos + 1 == len(row) or row[pos + 1] > 0):
                c[pos] = left_seven_eighths

        out.append("".join(c))

    return out



import sys

import pytest

import termplotlib as tpl


@pytest.mark.skipif(
    sys.stdout.encoding not in ["UTF-8", "UTF8"],
    reason=f"Need UTF-8 terminal (not {sys.stdout.encoding})",
)
def test_barh():
    fig = tpl.figure()
    fig.barh([3, 10, 5, 2], ["Cats", "Dogs", "Cows", "Geese"])
    # fig.show()
    string = fig.get_string()

    assert (
        string
        == """\
Cats   [ 3]  ████████████
Dogs   [10]  ████████████████████████████████████████
Cows   [ 5]  ████████████████████
Geese  [ 2]  ████████\
"""
    )
    return


def test_barh_ascii():
    fig = tpl.figure()
    fig.barh([3, 10, 5, 2], ["Cats", "Dogs", "Cows", "Geese"], force_ascii=True)
    # fig.show()
    string = fig.get_string()

    assert (
        string
        == """\
Cats   [ 3]  ************
Dogs   [10]  ****************************************
Cows   [ 5]  ********************
Geese  [ 2]  ********\
"""
    )
    return


@pytest.mark.skipif(
    sys.stdout.encoding not in ["UTF-8", "UTF8"],
    reason=f"Need UTF-8 terminal (not {sys.stdout.encoding})",
)
def test_barh_floats():
    fig = tpl.figure()
    fig.barh([0.3, 0.4, 0.6, 0.2], ["Cats", "Dogs", "Cows", "Geese"])
    # fig.show()
    string = fig.get_string()

    assert (
        string
        == """\
Cats   [0.3]  ████████████████████
Dogs   [0.4]  ██████████████████████████▋
Cows   [0.6]  ████████████████████████████████████████
Geese  [0.2]  █████████████▍\
"""
    )
    return


if __name__ == "__main__":
    # test_horizontal_ascii()
    # test_barh()
    test_barh_floats()



from .__about__ import __version__
from .figure import Figure, figure
from .subplot import SubplotGrid, subplot_grid

__all__ = [
    "__version__",
    "Figure",
    "figure",
    "SubplotGrid",
    "subplot_grid",
]



from .barh import barh
from .helpers import create_padding_tuple
from .hist import hist
from .plot import plot


def figure(*args, **kwargs):
    return Figure(*args, **kwargs)


class Figure:
    def __init__(self, width=None, padding=0):
        self._content = []
        self._width = width
        self._subfigures = None
        self._padding = create_padding_tuple(padding)
        return

    def aprint(self, string):
        self._content.append(string.split("\n"))
        return

    def show(self):
        print(self.get_string())
        return

    def get_string(self, remove_trailing_whitespace=True):
        lines = []

        padding_lr = self._padding[1] + self._padding[3]

        if self._width is None:
            width = max([len(line) for c in self._content for line in c])
            width += padding_lr
        else:
            width = self._width

        # Top padding
        lines += self._padding[0] * [" " * width]

        pr = " " * self._padding[1]
        pl = " " * self._padding[3]
        lines += [
            pl + line[: width - padding_lr] + pr for c in self._content for line in c
        ]

        # Bottom padding
        lines += self._padding[2] * [" " * width]

        if remove_trailing_whitespace:
            lines = [line.rstrip() for line in lines]

        return "\n".join(lines)

    def hist(self, *args, **kwargs):
        self._content.append(hist(*args, **kwargs))
        return

    def barh(self, *args, **kwargs):
        self._content.append(barh(*args, **kwargs))
        return

    def plot(self, *args, **kwargs):
        self._content.append(plot(*args, **kwargs))
        return



import sys


def create_padding_tuple(padding):
    # self._padding is a 4-tuple: top, right, bottom, left (just like CSS)
    if isinstance(padding, int):
        out = (padding, padding, padding, padding)
    else:
        if len(padding) == 1:
            out = (padding[0], padding[0], padding[0], padding[0])
        elif len(padding) == 2:
            out = (padding[0], padding[1], padding[0], padding[1])
        elif len(padding) == 3:
            out = (padding[0], padding[1], padding[2], padding[1])
        else:
            assert len(padding) == 4
            out = (padding[0], padding[1], padding[2], padding[3])
    return out


def is_unicode_standard_output():
    return hasattr(sys.stdout, "encoding") and sys.stdout.encoding.lower() in (
        "utf-8",
        "utf8",
    )



from .helpers import is_unicode_standard_output


def _trim_trailing_zeros(lst):
    k = 0
    for item in lst[::-1]:
        if item != 0:
            break
        k += 1
    return lst[:-k] if k > 0 else lst


def barh(
    vals, labels=None, max_width=40, bar_width=1, show_vals=True, force_ascii=False
):
    matrix = _get_matrix_of_eighths(vals, max_width, bar_width)

    if is_unicode_standard_output() and not force_ascii:
        chars = [" ", "▏", "▎", "▍", "▌", "▋", "▊", "▉", "█"]
    else:
        chars = [" ", "*", "*", "*", "*", "*", "*", "*", "*"]

    fmt = []
    if labels is not None:
        cfmt = "{{:{}s}}".format(max([len(str(label)) for label in labels]))
        fmt.append(cfmt)

    if show_vals:
        all_int = all(val == int(val) for val in vals)
        if all_int:
            cfmt = "{{:{}d}}".format(max([len(str(val)) for val in vals]))
        else:
            cfmt = "{}"
        fmt.append("[" + cfmt + "]")

    fmt.append("{}")
    fmt = "  ".join(fmt)

    out = []
    for k, (val, row) in enumerate(zip(vals, matrix)):
        data = []
        if labels is not None:
            data.append(str(labels[k]))
        if show_vals:
            data.append(val)

        # Cut off trailing zeros
        r = _trim_trailing_zeros(row)
        data.append("".join(chars[item] for item in r))
        out.append(fmt.format(*data))

    return out


def _get_matrix_of_eighths(counts, max_size, bar_width):
    max_count = max(counts)

    # translate to eighths of a textbox
    eighths = [int(round(count / max_count * max_size * 8)) for count in counts]

    # prepare matrix
    matrix = [[0] * max_size for _ in range(len(eighths))]
    for i, eighth in enumerate(eighths):
        num_full_blocks = eighth // 8
        remainder = eighth % 8
        for j in range(num_full_blocks):
            matrix[i][j] = 8
        if remainder > 0:
            matrix[i][num_full_blocks] = remainder

    # Account for bar width
    out = []
    for i in range(len(matrix)):
        for _ in range(bar_width):
            out.append(matrix[i])
    return out



import termplotlib as tpl


def _generate_content(*args, **kwargs):
    grid = tpl.subplot_grid((2, 3), *args, **kwargs)
    grid[0, 0].aprint("Some text")
    grid[0, 1].aprint("Some more text\nand more")
    grid[0, 2].aprint("Some more text\nand more\neven more")
    grid[1, 0].aprint("Some more text\nand more\neven more")
    grid[1, 1].aprint("Some more text\nand more")
    grid[1, 2].aprint("Some text")
    return grid


def test_subplot():
    grid = tpl.subplot_grid((1, 2), width=20)
    grid[0, 0].aprint("Some text")
    grid[0, 1].aprint("Some more text")
    string = grid.get_string()
    assert (
        string
        == """┌─────────┬────────┐
│         │        │
│  Some   │  Some  │
│         │        │
└─────────┴────────┘"""
    )
    return


def test_subplot_custom_border():
    grid = tpl.subplot_grid((1, 2), border_style="x", width=20)
    grid[0, 0].aprint("Some text")
    grid[0, 1].aprint("Some more text")
    string = grid.get_string()
    assert (
        string
        == """xxxxxxxxxxxxxxxxxxxx
x         x        x
x  Some   x  Some  x
x         x        x
xxxxxxxxxxxxxxxxxxxx"""
    )
    return


def test_subplot2():
    grid = tpl.subplot_grid((1, 2), width=20, border_style="thick")
    grid[0, 0].aprint("Some text")
    grid[0, 1].aprint("Some more text\nand more")
    string = grid.get_string()
    assert (
        string
        == """┏━━━━━━━━━┳━━━━━━━━┓
┃         ┃        ┃
┃  Some   ┃  Some  ┃
┃         ┃  and   ┃
┃         ┃        ┃
┗━━━━━━━━━┻━━━━━━━━┛"""
    )
    return


def test_subplot_padding():
    grid = tpl.subplot_grid((1, 2), width=20, padding=2, border_style="double")
    grid[0, 0].aprint("Some text")
    grid[0, 1].aprint("Some more text\nand more")
    string = grid.get_string()
    assert (
        string
        == """╔═════════╦════════╗
║         ║        ║
║         ║        ║
║  Some   ║  Some  ║
║         ║  and   ║
║         ║        ║
║         ║        ║
╚═════════╩════════╝"""
    )
    return


def test_subplot_3x2():
    grid = _generate_content(width=40)
    string = grid.get_string()

    assert (
        string
        == """┌────────────┬────────────┬────────────┐
│            │            │            │
│  Some tex  │  Some mor  │  Some mor  │
│            │  and more  │  and more  │
│            │            │  even mor  │
│            │            │            │
├────────────┼────────────┼────────────┤
│            │            │            │
│  Some mor  │  Some mor  │  Some tex  │
│  and more  │  and more  │            │
│  even mor  │            │            │
│            │            │            │
└────────────┴────────────┴────────────┘"""
    )
    return


def test_subplot_ascii():
    grid = _generate_content(width=40, border_style="ascii")
    string = grid.get_string()

    assert (
        string
        == """----------------------------------------
|            |            |            |
|  Some tex  |  Some mor  |  Some mor  |
|            |  and more  |  and more  |
|            |            |  even mor  |
|            |            |            |
|------------+------------+------------|
|            |            |            |
|  Some mor  |  Some mor  |  Some tex  |
|  and more  |  and more  |            |
|  even mor  |            |            |
|            |            |            |
----------------------------------------"""
    )
    return


def test_subplot_thin_rounded():
    grid = _generate_content(width=40, border_style="thin rounded")
    string = grid.get_string()

    assert (
        string
        == """╭────────────┬────────────┬────────────╮
│            │            │            │
│  Some tex  │  Some mor  │  Some mor  │
│            │  and more  │  and more  │
│            │            │  even mor  │
│            │            │            │
├────────────┼────────────┼────────────┤
│            │            │            │
│  Some mor  │  Some mor  │  Some tex  │
│  and more  │  and more  │            │
│  even mor  │            │            │
│            │            │            │
╰────────────┴────────────┴────────────╯"""
    )
    return


def test_subplot_custom():
    grid = _generate_content(
        width=40, border_style=["-", "|", "-", "-", "-", "-", "|", "|", "T", "-", "X"]
    )
    string = grid.get_string()

    assert (
        string
        == """-------------T------------T-------------
|            |            |            |
|  Some tex  |  Some mor  |  Some mor  |
|            |  and more  |  and more  |
|            |            |  even mor  |
|            |            |            |
|------------X------------X------------|
|            |            |            |
|  Some mor  |  Some mor  |  Some tex  |
|  and more  |  and more  |            |
|  even mor  |            |            |
|            |            |            |
----------------------------------------"""
    )

    grid.show()
    return


def test_subplot_no_borders():
    grid = _generate_content(width=40, border_style=None)
    string = grid.get_string()

    ref = """
  Some tex    Some mor    Some mor
              and more    and more
                          even mor


  Some mor    Some mor    Some tex
  and more    and more
  even mor"""

    assert string == ref
    return


def test_subplot_column_widths():
    grid = _generate_content(column_widths=(30, 15, 20))
    string = grid.get_string()

    print(string)

    assert (
        string
        == """┌──────────────────────────────┬───────────────┬────────────────────┐
│                              │               │                    │
│  Some text                   │  Some more t  │  Some more text    │
│                              │  and more     │  and more          │
│                              │               │  even more         │
│                              │               │                    │
├──────────────────────────────┼───────────────┼────────────────────┤
│                              │               │                    │
│  Some more text              │  Some more t  │  Some text         │
│  and more                    │  and more     │                    │
│  even more                   │               │                    │
│                              │               │                    │
└──────────────────────────────┴───────────────┴────────────────────┘"""
    )
    return


if __name__ == "__main__":
    test_subplot_padding()



import subprocess


def plot(
    x,
    y,
    width=80,
    height=25,
    label=None,
    xlim=None,
    ylim=None,
    xlabel=None,
    title=None,
    extra_gnuplot_arguments=None,
    plot_command="plot '-' w lines",
    ticks_scale=0,
):
    p = subprocess.Popen(
        ["gnuplot"],
        stdout=subprocess.PIPE,
        stdin=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    gnuplot_input = []

    gnuplot_input.append(f"set term dumb mono {width},{height}")
    # gnuplot_input.append("set tics nomirror")
    gnuplot_input.append(f"set tics scale {ticks_scale}")

    if xlim:
        gnuplot_input.append("set xrange [{}:{}]".format(xlim[0], xlim[1]))

    if ylim:
        gnuplot_input.append("set yrange [{}:{}]".format(ylim[0], ylim[1]))

    if xlabel:
        gnuplot_input.append(f'set xlabel "{xlabel}"')

    if title:
        gnuplot_input.append(f'set title "{title}"')

    if extra_gnuplot_arguments:
        gnuplot_input += extra_gnuplot_arguments

    string = plot_command
    if label:
        string += f" title '{label}'"
    else:
        string += " notitle"

    gnuplot_input.append(string)

    for xx, yy in zip(x, y):
        gnuplot_input.append(f"{xx:e} {yy:e}")
    gnuplot_input.append("e")

    out = p.communicate(input="\n".join(gnuplot_input).encode())[0]

    return _remove_empty_lines(out.decode())


def _remove_empty_lines(string):
    return string.split("\n")[1:-2]
